import asyncio
import logging
import mimetypes
import traceback
from typing import Coroutine

from watchdog.events import FileModifiedEvent
from watchdog.observers import Observer

from dtps import DTPSContext
from dtps_http import RawData

# suppress watchdog logging
for logger in [logging.getLogger(name) for name in logging.root.manager.loggerDict]:
    if logger.name.startswith("watchdog.observers."):
        logger.setLevel(logging.WARNING)


async def task(coro: Coroutine):
    # noinspection PyBroadException
    try:
        await coro
    except Exception:
        print(f"Exception in task: {coro.__name__}")
        traceback.print_exc()


class FileModifiedEventHandler(FileModifiedEvent):

    def __init__(self, path: str, cxt: DTPSContext, loop: asyncio.AbstractEventLoop):
        super().__init__(path)
        self._cxt: DTPSContext = cxt
        self._loop: asyncio.AbstractEventLoop = loop

    def dispatch(self, event):
        if event.event_type != "closed":
            return
        # re-read the file and re-publish it

        async def republish():
            try:
                with open(self.src_path, "rb") as f:
                    data: bytes = f.read()
                    mime: str = mimetypes.guess_type(self.src_path)[0]
                    await self._cxt.publish(RawData(content=data, content_type=mime or "application/octet-stream"))
                    print(f"File {self.src_path} reloaded")
            except Exception as e:
                print(f"Error reading file: {e}")

        self._loop.create_task(task(republish()))


def setup_file_watcher(path: str, cxt: DTPSContext, loop: asyncio.AbstractEventLoop):
    handler = FileModifiedEventHandler(path, cxt, loop)
    observer = Observer()
    observer.schedule(handler, path, recursive=True)
    observer.start()



import dataclasses
import os
from types import NoneType
from typing import Union, Literal, TYPE_CHECKING, Callable, Awaitable, Dict, Tuple

from aiohttp import web
from jinja2 import Template

from dtps_http import RawData, HTTPResponse

Response = Union[RawData, HTTPResponse, 'HTML']
EventType = Literal["register", "unregister", "update", "style", "action"]
ValueType = Union[int, float, str, list, dict, None]
AllowedValueTypes = (int, float, str, list, dict, NoneType)
ValuePattern = str

if TYPE_CHECKING:
    import dtps_ui


@dataclasses.dataclass
class Event:
    type: EventType
    selector: str
    key: str
    value: ValueType

    def to_dict(self) -> dict:
        return dataclasses.asdict(self)


class Style:
    """
    The style of an element.
    """
    __cache__: Dict[Tuple[int, str, str], 'Style'] = {}

    def __init__(self, app: 'dtps_ui.App', path: str, selector: str):
        cache_key = (id(app), path, selector)
        if cache_key in Style.__cache__:
            raise ValueError(f"Style already exists: {cache_key}, use Style.get() instead.")
        self._app = app
        self._path = path
        self._selector = selector

    @staticmethod
    def get(app: 'dtps_ui.App', path: str, selector: str) -> 'Style':
        cache_key = (id(app), path, selector)
        if cache_key in Style.__cache__:
            return Style.__cache__[cache_key]
        elem = Style(app, path, selector)
        Style.__cache__[cache_key] = elem
        return elem

    def __setattr__(self, key, value):
        if key.startswith("_"):
            super().__setattr__(key, value)
        else:
            if not isinstance(value, AllowedValueTypes):
                raise ValueError(f"Value type not allowed: '{type(value)}'; "
                                 f"Allowed types: {', '.join(map(str, AllowedValueTypes))}")
            self._app.publish_events(self._path, [Event(type="style", selector=self._selector, key=key, value=value)])


class Element:
    """
    Any element that can generate and react to events.
    """
    __cache__: Dict[Tuple[int, str, str], 'Element'] = {}

    def __init__(self, app: 'dtps_ui.App', path: str, selector: str):
        cache_key = (id(app), path, selector)
        if cache_key in Element.__cache__:
            raise ValueError(f"Element already exists: {cache_key}, use Element.get() instead.")
        self._app = app
        self._path = path
        self._selector = selector

    @staticmethod
    def get(app: 'dtps_ui.App', path: str, selector: str) -> 'Element':
        cache_key = (id(app), path, selector)
        if cache_key in Element.__cache__:
            return Element.__cache__[cache_key]
        elem = Element(app, path, selector)
        Element.__cache__[cache_key] = elem
        return elem

    @property
    def style(self) -> Style:
        return Style.get(self._app, self._path, self._selector)

    def on(self, key: str, handler: Callable[[Event], Awaitable], value: ValuePattern = None) -> None:
        """
        Register a handler for an event.
        """
        # register listener with the app
        self._app.listen_for(self._path, self._selector, "action", key, value, handler)
        # register listener on the frontend
        self._app.publish_events(self._path, [Event(type="register", selector=self._selector, key=key, value=value)])

    def __setattr__(self, key, value):
        if key.startswith("_"):
            super().__setattr__(key, value)
        else:
            if not isinstance(value, AllowedValueTypes):
                raise ValueError(f"Value type not allowed: '{type(value)}'; "
                                 f"Allowed types: {', '.join(map(str, AllowedValueTypes))}")
            self._app.publish_events(self._path, [Event(type="update", selector=self._selector, key=key, value=value)])


@dataclasses.dataclass
class HTML:
    content: str

    @classmethod
    def from_file(cls, path: str) -> 'HTML':
        return cls.from_template(path, None)

    @classmethod
    def from_template(cls, path: str, context: dict = None) -> 'HTML':
        content: str
        with open(path, "rt") as f:
            content = f.read()
        # apply context using jinja2
        context = context.copy() if context else {}
        # add dtps include html
        dtps_ui_include_fpath = os.path.join(os.path.dirname(__file__), "assets", "dtps_ui_include.html")
        with open(dtps_ui_include_fpath, "rt") as f:
            dtps_ui_include_html = f.read()
        context["dtps_ui_include"] = dtps_ui_include_html
        # render template
        template = Template(content)
        content = template.render(**context)
        # return HTML
        return HTML(content=content)

    def to_response(self) -> web.Response:
        return web.Response(body=self.content, content_type="text/html")


class FrontEnd:
    __cache__: Dict[Tuple[int, str], 'FrontEnd'] = {}

    def __init__(self, app: 'dtps_ui.App', path: str):
        cache_key = (id(app), path)
        if cache_key in FrontEnd.__cache__:
            raise ValueError(f"FrontEnd already exists: {cache_key}, use FrontEnd.get() instead.")
        self._app = app
        self._path = path

    @staticmethod
    def get(app: 'dtps_ui.App', path: str):
        cache_key = (id(app), path)
        if cache_key in FrontEnd.__cache__:
            return FrontEnd.__cache__[cache_key]
        fe = FrontEnd(app, path)
        FrontEnd.__cache__[cache_key] = fe
        return fe

    @property
    def document(self) -> Element:
        """
        Stub for the document element.
        """
        return Element.get(self._app, self._path, "document")

    @property
    def window(self) -> Element:
        """
        Stub for the window element.
        """
        return Element.get(self._app, self._path, "window")

    def element(self, selector: str) -> Element:
        """
        Stub for an element.
        """
        return Element.get(self._app, self._path, selector)



import logging
import os.path
from typing import Optional

import argparse

from dtps_http import HTTPRequest
from dtps_ui import App
from dtps_ui.types import HTML

DEFAULT_STATIC_DIR = "./static"

logging.basicConfig()
logger = logging.getLogger("dtps-ui-server")
logger.setLevel(logging.INFO)


def default_static_dir() -> Optional[str]:
    return DEFAULT_STATIC_DIR if os.path.isdir(DEFAULT_STATIC_DIR) else None


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--host", type=str, default="localhost", help="Host to bind to")
    parser.add_argument("--port", type=int, default=8000, help="Port to bind to")
    parser.add_argument("--root", type=str, default="/app/", help="Root URL")
    parser.add_argument("--static-dir", type=str, default=default_static_dir(), help="Static files directory")
    parser.add_argument("--log-level", type=str, default="INFO", help="Log level")
    parser.add_argument("html", type=str, help="HTML file to serve")

    args = parser.parse_args()

    logger.setLevel(args.log_level)

    app = App(host=args.host, port=args.port, root=args.root, static_dirs=args.static_dir)

    @app.route("/")
    async def index(request: HTTPRequest) -> HTML:
        logger.debug(f"GET: {request.url}")
        return HTML.from_file(args.html)

    app.serve_forever()


if __name__ == "__main__":
    main()



import asyncio
import base64
import contextlib
import glob
import json
import mimetypes
import os
from asyncio import AbstractEventLoop
from collections import defaultdict
from threading import RLock
from typing import Optional, Dict, Callable, Awaitable, Iterable, List, Tuple, cast, Union

from dtps import DTPSContext, context as dtps_context
from dtps_http import RawData, HTTPRequest
from .types import Event, Response, HTML, FrontEnd, EventType, ValuePattern
from .utils import task, setup_file_watcher


Path = str
Selector = str
Key = str


class EventsQueue:

    def __init__(self, state_cxt: DTPSContext, events_cxt: DTPSContext, loop: AbstractEventLoop):
        self._buffer: Optional[List[Event]] = None
        self._buffer_lock: RLock = RLock()
        self._state_cxt: DTPSContext = state_cxt
        self._events_cxt: DTPSContext = events_cxt
        self._loop: AbstractEventLoop = loop

    async def publish(self, evts: Iterable[Event]):
        with self._buffer_lock:
            if self._buffer is None:
                # publish the events immediately
                await self.do_publish(evts)
            else:
                # buffer the events for later
                self._buffer.extend(evts)

    async def do_publish(self, evts: Iterable[Event]):
        # update the state
        patch = []
        for evt in evts:
            path: str = base64.b64encode(f"/{evt.type}/{evt.selector}/{evt.key}".encode("utf-8")).decode("utf-8")
            patch.append({"op": "add", "path": f"/{path}", "value": evt.to_dict()})
        await self._state_cxt.patch(patch)
        # publish out the events
        evts_json: str = json.dumps([evt.to_dict() for evt in evts])
        evts_rd: RawData = RawData(content=evts_json.encode("utf-8"), content_type="application/json")
        await self._events_cxt.publish(evts_rd)

    @contextlib.contextmanager
    async def atomic(self):
        with self._buffer_lock:
            self._buffer = []
        # return control to the caller
        yield
        # get the events collected
        with self._buffer_lock:
            evts = self._buffer
            self._buffer = None
        # publish the events
        await self.do_publish(evts)


class App:

    def __init__(self, host: str, port: int, root: str = None, static_dirs: Union[str, Iterable[str]] = None,
                 loop: AbstractEventLoop = None):
        self._host = host
        self._port = port
        self._root = (root or "").strip("/")
        self._static_dirs: Iterable[str] = \
            (static_dirs if isinstance(static_dirs, list) else [static_dirs]) if static_dirs else None
        self._root_cxt: Optional[DTPSContext] = None
        self._loop = loop or asyncio.get_event_loop()
        self._routes: Dict[Path, DTPSContext] = {}
        self._event_queues: Dict[Path, EventsQueue] = {}
        self._event_listeners: Dict[Tuple[Path, Selector, EventType, Key], List[Callable[[Event], Awaitable]]] = \
            defaultdict(list)
        self._ready: asyncio.Event = asyncio.Event()

    async def astart(self):
        self._loop = asyncio.get_event_loop()
        await self._init()
        self._ready.set()
        await self._run()

    async def _init(self):
        url0: str = f"create:http://{self._host}:{self._port}/"
        self._root_cxt = (await dtps_context("app", urls=[url0])).navigate(self._root)

    async def _run(self):
        # load static files
        if self._static_dirs is not None:
            for static_dir in self._static_dirs:
                for file in glob.glob(os.path.join(static_dir, "**", "*"), recursive=True):
                    rel_path: str = os.path.relpath(file, static_dir)
                    route: DTPSContext = await (self._root_cxt / "static" / rel_path).queue_create()
                    self._routes[rel_path] = route
                    with open(file, "rb") as f:
                        data: bytes = f.read()
                        mime: str = mimetypes.guess_type(file)[0]
                        await route.publish(RawData(content=data, content_type=mime or "application/octet-stream"))
                    setup_file_watcher(file, route, self._loop)
        # load dtps_ui assets
        assets_dir: str = os.path.join(os.path.dirname(__file__), "assets")
        # noinspection PyUnresolvedReferences
        for file in glob.glob(os.path.join(assets_dir, "**", "*"), recursive=True):
            rel_path: str = os.path.relpath(file, assets_dir)
            route: DTPSContext = await (self._root_cxt / "assets" / rel_path).queue_create()
            self._routes[rel_path] = route
            with open(file, "rb") as f:
                data: bytes = f.read()
                mime: str = mimetypes.guess_type(file)[0]
                await route.publish(RawData(content=data, content_type=mime or "application/octet-stream"))
            setup_file_watcher(file, route, self._loop)
        # wait until the app is shutdown
        try:
            while True:
                await asyncio.sleep(1)
        except asyncio.CancelledError:
            print("Event loop stopping...")

    async def event_queue(self, path: str) -> EventsQueue:
        await self._ready.wait()
        evt_queue: EventsQueue = self._event_queues.get(path, None)
        if evt_queue is not None:
            return evt_queue
        _path: str = path.strip("/")
        # create empty events queue
        state_cxt = await (self._root_cxt / _path / "__state__").queue_create()
        await state_cxt.publish(RawData.cbor_from_native_object({}))
        # create empty events queues
        our_events_cxt = await (self._root_cxt / _path / "__frombackend__").queue_create()
        await our_events_cxt.publish(RawData.cbor_from_native_object([]))
        their_events_cxt = await (self._root_cxt / _path / "__tobackend__").queue_create()
        await their_events_cxt.publish(RawData.cbor_from_native_object([]))

        async def _process_events(rd: RawData):
            evts = [
                Event(**evt) for evt in cast(list, rd.get_as_native_object())
            ]
            await self._process_events(path, evts)
        await their_events_cxt.subscribe(_process_events)
        # create the events queue handler object
        events_queue = EventsQueue(state_cxt, our_events_cxt, self._loop)
        self._event_queues[path] = events_queue
        return events_queue

    def publish_events(self, path: Path, evts: Iterable[Event]):
        self._loop.create_task(task(self.apublish_events(path, evts)))

    async def apublish_events(self, path: Path, evts: Iterable[Event]):
        evt_queue: EventsQueue = await self.event_queue(path)
        await evt_queue.publish(evts)

    def listen_for(self, path: Path, selector: Selector, event: EventType, key: str, value: ValuePattern,
                   handler: Callable[[Event], Awaitable]) -> None:
        # return key
        rkey: str = f"{key}|{value or ''}"
        self._event_listeners[(path, selector, event, rkey)].append(handler)

    async def _process_events(self, path: Path, evts: Iterable[Event]):
        for evt in evts:
            handlers: List[Callable[[Event], Awaitable]] = self._event_listeners.get(
                (path, evt.selector, evt.type, evt.key), []
            )
            for handler in handlers:
                await handler(evt)

    async def atomic(self, path: Path):
        evt_queue: EventsQueue = await self.event_queue(path)
        return evt_queue.atomic

    def route(self, path: Path):
        if path in self._routes:
            raise ValueError(f"Route {path} already exists")

        def wrapper(f: Callable[[HTTPRequest], Awaitable[Response]]):
            async def _setup_route():
                await self._ready.wait()

                # create empty events queue
                await self.event_queue(path)

                async def serve(request: HTTPRequest) -> Response:
                    rs = await f(request)
                    if isinstance(rs, HTML):
                        rs = rs.to_response()
                    return rs

                _path: str = path.strip("/")
                route: DTPSContext = await (self._root_cxt / _path).queue_create(serve=serve)
                self._routes[path] = route

            self._loop.create_task(task(_setup_route()))

        return wrapper

    def background(self, f: Callable[[], Awaitable[None]]):
        async def _setup_background_task():
            # wait until the app is ready
            await self._ready.wait()
            # run background task
            await f()

        self._loop.create_task(task(_setup_background_task()))

    def frontend(self, path: Path) -> FrontEnd:
        return FrontEnd.get(self, path)

    def serve_forever(self):
        try:
            if self._loop.is_running():
                raise ValueError("The event loop is already running. Use the async entrypoint `App.arun()` instead.")
            self._loop.run_until_complete(self.astart())
        except KeyboardInterrupt:
            pass



from setuptools import find_packages, setup

# :==> Fill in your project data here
package_name = 'dtps-ui'
library_webpage = 'https://github.com/duckietown/lib-dtps-ui'
maintainer = 'Andrea F. Daniele'
maintainer_email = 'afdaniele@duckietown.com'
short_description = 'Use DTPS to create dynamic web apps.'
full_description = """
Use DTPS to create dynamic web apps.
"""

if "<" in package_name:
    msg = "Please fill in the project data in setup.py."
    raise ValueError(msg)


# Read version from the __init__ file
def get_version_from_source(filename):
    import ast
    v = None
    with open(filename) as f:
        for line in f:
            if line.startswith("__version__"):
                v = ast.parse(line).body[0].value.s
                break
        else:
            raise ValueError("No version found in %r." % filename)
    if v is None:
        raise ValueError(filename)
    return v


version = get_version_from_source("src/dtps_ui/__init__.py")

# read project dependencies
# NO - dependencies.txt is for testing dependiences - EVERYTHING PINNED
# The requirements here must be broad.
# dependencies_file = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'dependencies.txt')
# with open(dependencies_file, 'rt') as fin:
#     dependencies = list(filter(lambda line: not line.startswith('#'), fin.read().splitlines()))

install_requires = [
    "dtps_http",
    "jinja2",
    "watchdog",
]
tests_require = []

# compile description
underline = "=" * (len(package_name) + len(short_description) + 2)
description = """
{name}: {short}
{underline}

{long}
""".format(
    name=package_name,
    short=short_description,
    long=full_description,
    underline=underline,
)

console_scripts = [
    "dtps-ui-serve = dtps_ui.cli_serve:main",
]

# setup package
setup(
    name=package_name,
    author=maintainer,
    author_email=maintainer_email,
    url=library_webpage,
    tests_require=tests_require,
    install_requires=install_requires,
    package_dir={"": "src"},
    packages=find_packages("./src"),
    package_data={
        'dtps_ui': [
            'assets/*',
            'assets/*/*',
            'assets/*/*/*',
            'assets/*/*/*/*',
        ],
    },
    include_package_data=True,
    long_description=description,
    version=version,
    entry_points={"console_scripts": console_scripts},
)
